module wei_kxmig
  !! fk-fx migration
  !! receive wfld at (izs,iws,ith)
  !! return  wfld downward/upward continued
  !! needs FKop, FXop

  use wei_util
  use wei_run
  use wei_external
  use wei_down
  use wei_obj_S
  use wei_fft
  use wei_kmap
  !---------------
  implicit none

  !! temporary interpolation arrays
  complex, dimension(:,:,:,:,:), pointer, private :: tmp1,tmp2
  complex, dimension(:,:), pointer, private :: tmps
  character(len=128) :: twave

contains
  !----------------------------------------------------------------
  !----------------------------------------------------------------
  subroutine weikxmig(wfld,iws,izs,ith,KXMIGop,FKop,FXop)
    complex, dimension(:,:,:,:,:), pointer :: wfld
    integer,external,optional ::    FKop,FXop
    integer :: izs,iws,ith
    integer :: st
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmig'
#endif
    interface !!-------------------------
       integer function KXMIGop(wfld,iws,izs,ith,FKop,FXop) result(st)
         complex, dimension(:,:,:,:,:), pointer :: wfld
         integer, intent(in)       :: iws,izs,ith
         integer,external,optional :: FKop,FXop
       end function KXMIGop
    end interface !!---------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

    st = KXMIGop(wfld,iws,izs,ith,FKop,FXop)

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine weikxmig
  !----------------------------------------------------------------
!!$
!!$=head1 NAME
!!$
!!$ weikxmig1_init - Single velocity KX-migration initialization
!!$
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr= weikxmig1_init()
!!$
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Initialize a single-reference kx-migration velocity operation
!!$
!!$
!!$=head1 SEE ALSO
!!$
!!$L<weikxmig1> L<weikxmigN_init>
!!$
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$
!!$=cut
!!$
  !----------------------------------------------------------------
        integer function weikxmig1_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmig1_init'
    if(run%debug) call in(unit)
#endif
    !! do nothing
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weikxmig1_init
!!$
!!$=head1 NAME
!!$
!!$ weikxmig1 - Single velocity KX-migration 
!!$
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr= weikxmig1(wfld,iws,izs,ith,FKOP,FXOP)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item wfld complex(:,:,:,:,:)
!!$
!!$  Wavefield
!!$
!!$=item iws - integer
!!$
!!$ Current frequency slice
!!$
!!$=item izs
!!$
!!$ Current depth slice
!!$
!!$=item ith
!!$
!!$ Current thread
!!$
!!$=item FKOP FK interface L<weiwem>
!!$
!!$=item FXOP FX interface L<weissf>
!!$
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Perform  a single-reference kx-migration velocity operation
!!$
!!$          st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op
!!$
!!$          call weiift(iws,        ith,wfld)
!!$
!!$          st   = FKop(iws,izs,ifk,ith,wfld) !! f-k op
!!$
!!$          call weifft(iws,        ith,wfld)
!!$
!!$
!!$=head1 SEE ALSO
!!$
!!$=head1 SEE ALSO
!!$
!!$L<weikxmigN> L<weikxmig1_init>
!!$
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$
!!$=cut
!!$
  !----------------------------------------------------------------
  !! 1 reference slowness
  integer function weikxmig1(wfld,iws,izs,ith,FKop,FXop) result(st)
    complex, dimension(:,:,:,:,:), pointer :: wfld
    integer, intent(in)       :: iws,izs,ith
    integer                   :: ifk
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmig1'
#endif
    interface !!----------------------------------------------
       integer function FKop(iws,izs,ifk,ith,wfld)      result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FKop
       integer function FXop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FXop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

    call redo_kmaps(var%ihxb_beg(izs),var%ihxb_end(izs), &
       var%ihyb_beg(izs),var%ihyb_end(izs))

    if(down%adj) then !! MIGRATION
       st   = FXop(iws,izs,  1,ith,wfld) !! f-x op
       call weiift(iws,izs,    ith,wfld)
       st   = FKop(iws,izs,  1,ith,wfld) !! f-k op
       call weifft(iws,izs,    ith,wfld)
    else             !! MODELLING
       call weiift(iws,izs,    ith,wfld)
       st   = FKop(iws,izs,  1,ith,wfld) !! f-k op
       call weifft(iws,izs,    ith,wfld)
       st   = FXop(iws,izs,  1,ith,wfld) !! f-x op
    end if

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weikxmig1
!!$
!!$=head1 NAME
!!$
!!$ weikxmigN_init - Multi-velocity KX-migration initialization
!!$
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr= weikxmigN_init()
!!$
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Initialize a multi-reference kx-migration velocity operation
!!$
!!$
!!$=head1 SEE ALSO
!!$
!!$L<weikxmigN> L<weikxmig1_init>
!!$
!!$
!!$=head1 sepweif90
!!$
!!$ sepweif90
!!$
!!$
!!$=cut
!!$
  !----------------------------------------------------------------
  !! N reference slownesses
  integer function weikxmigN_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmigN_init'
    if(run%debug) call in(unit)
#endif

      call init_wave_temps()
!      !! allocate tmp arrays for interpolation of multiple fk ops
!      allocate(tmp1(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,run%o_nth),stat=er)
!      if(er/=0) call seperr('cant allocate tmp1')
!      allocate(tmp2(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,run%o_nth),stat=er)
!      if(er/=0) call seperr('cant allocate tmp2')

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weikxmigN_init
!!$
!!$=head1 NAME
!!$
!!$ weikxmigN - Multi- velocity KX-migration 
!!$
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr= weikxmigN(wfld,iws,izs,ith,FKOP,FXOP)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item wfld complex(:,:,:,:,:)
!!$
!!$  Wavefield
!!$
!!$=item iws - integer
!!$
!!$ Current frequency slice
!!$
!!$=item izs
!!$
!!$ Current depth slice
!!$
!!$=item ith
!!$
!!$ Current thread
!!$
!!$=item FKOP FK interface L<weiwem>
!!$
!!$=item FXOP FX interface L<weissf>
!!$
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Perform  a single-reference kx-migration velocity operation
!!$
!!$          st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op
!!$
!!$          call weiift(iws,izs, izs,ith,wfld)
!!$
!!$          st   = FKop(iws,izs,ifk,ith,wfld) !! f-k op
!!$
!!$          call weifft(iws,izs,ith,wfld)
!!$
!!$
!!$=head1 SEE ALSO
!!$
!!$L<weikxmig1> L<weikxmigN_init>
!!$
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$
!!$=cut
!!$
  !----------------------------------------------------------------
  !! N reference slownesses
  integer function weikxmigN(wfld,iws,izs,ith,FKop,FXop) result(st)
    complex, dimension(:,:,:,:,:), pointer :: wfld
    integer, intent(in)       :: iws,izs,ith
    integer                   :: ifk
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmigN'
#endif
    interface !!----------------------------------------------
       integer function FKop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FKop
       integer function FXop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FXop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif
    if(down%adj) then !! MIGRATION
       !----------------------------------------------------------------       



       call store_wavefield(wfld,ith,iws)
!       tmp2(:,:,:,:,ith) = wfld(:,:,:,:,iws)
!       tmp1(:,:,:,:,ith) = 0.

        call redo_kmaps(var%ihxb_beg(izs),var%ihxb_end(izs), &
       var%ihyb_beg(izs),var%ihyb_end(izs))


   do ifk=1,slo%nrs(izs)
          

          call grab_part_wvfld(wfld,ifk,ith,iws,izs)
!          where(slo%si(:,:,:,:,izs) == ifk)
!             wfld(:,:,:,:,iws) = tmp2(:,:,:,:,ith)
!          elsewhere
!             wfld(:,:,:,:,iws) = 0.
!          end where

          st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op
          call weiift(iws,izs,    ith,wfld)
          st   = FKop(iws,izs,ifk,ith,wfld) !! f-k op
          call weifft(iws,izs,    ith,wfld)


          call add_wavefield(wfld,ith,iws)

!          tmp1(:,:,:,:,ith) = tmp1(:,:,:,:,ith) +  wfld(:,:,:,:,iws)

       end do
        call grab_wavefield(wfld,iws,ith)
!       wfld(:,:,:,:,iws) = tmp1(:,:,:,:,ith)
    



       !----------------------------------------------------------------
    else             !! MODELLING
       !----------------------------------------------------------------
      
!       tmp2(:,:,:,:,ith) = wfld(:,:,:,:,iws)
!       tmp1(:,:,:,:,ith) = 0.
        call store_wavefield(wfld,iws,ith)
   do ifk=1,slo%nrs(izs)
          call grab_wavefield(wfld,iws,ith)
!          wfld(:,:,:,:,iws) = tmp2(:,:,:,:,ith)

          call weiift(iws,izs,    ith,wfld)
          st   = FKop(iws,izs,ifk,ith,wfld) !! f-k op
          call weifft(iws,izs,    ith,wfld)
          st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op

          call store_part_wavefield(wfld,ifk,ith,iws,izs) 
          
!          where(slo%si(:,:,:,:,izs) == ifk)
!             tmp1(:,:,:,:,ith) = &
!             tmp1(:,:,:,:,ith) + &
!             wfld(:,:,:,:,iws)
!          end where

       end do
        call grab_wavefield(wfld,iws,ith)
!       wfld(:,:,:,:,iws) = tmp1(:,:,:,:,ith)
       !----------------------------------------------------------------
    end if

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weikxmigN
  !----------------------------------------------------------------  


!! multiple fk ops
!!
!!       D(z)
!!       |
!!       v
!!       T1
!!
!!    D1 D2 D3 ...
!!
!!    S1 S2 S3 ...
!!     \ | /
!!       T2
!!       |
!!       v
!!       D(z+dz)







!---------------------------------------------
!
!
!  Initialize storage for wavefields
!
!
subroutine init_wave_temps() 
 integer ::er


if(.not. run%mem_save) then

   !! allocate tmp arrays for interpolation of multiple fk ops
   allocate(tmp1(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,run%o_nth),stat=er)
   if(er/=0) call seperr('cant allocate tmp1')
   allocate(tmp2(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,run%o_nth),stat=er)
   if(er/=0) call seperr('cant allocate tmp2')
else 
   allocate(tmps(down%amx%n,down%amy%n))
   write(twave,"(a,a)") "T_WAVE",run%tag_num
   call auxtmp(twave)
end if
end subroutine

!-------------------------------------------------
!
!
! Store the wavefield
!
!
subroutine store_wavefield(wfld,ith,iws)
 complex :: wfld(:,:,:,:,:)
 integer :: ierr,iws,ith
  
if(run%mem_save) then
  ierr= sseek(twave,0,0)
  ierr= srite(twave,wfld(:,:,:,:,iws),4*size(wfld(:,:,:,:,iws)))
else
  tmp2(:,:,:,:,ith) = wfld(:,:,:,:,iws)
  tmp1=0.
end if
end subroutine

!------------------------------------------------------
!
!
!
!
subroutine grab_part_wvfld(wfld,ifk,ith,iws,izs)
integer :: ith,iws,ifk,ierr,izs
 complex :: wfld(:,:,:,:,:)

if(run%mem_save) then
  ierr= sseek(twave,0,0)
  ierr= sreed(twave,wfld(:,:,:,:,iws),4*size(wfld(:,:,:,:,iws)))
  call partition_slow(wfld(:,:,:,:,iws),ifk,izs)
else
     where(slo%si(:,:,:,:,izs) == ifk)
       wfld(:,:,:,:,iws) = tmp2(:,:,:,:,ith)
     elsewhere
         wfld(:,:,:,:,iws) = 0.
     end where
end if
end subroutine
!------------------------------------------------------
!
!
!
!
subroutine add_wavefield(wfld,ith,iws)
complex :: wfld(:,:,:,:,:)
integer :: ith,iws,i1,i2,ierr

if(run%mem_save) then
  ierr= sseek(twave,0,0)
   do i2=1,size(wfld,4)
     do i1=1,size(wfld,3)
       ierr= sreed(twave,tmps,size(tmps)*4)
       wfld(:,:,i1,i2,iws)=wfld(:,:,i1,i2,iws)+tmps
     end do
   end do 
  ierr= sseek(twave,0,0)
  ierr= srite(twave,wfld(:,:,:,:,iws),4*size(wfld(:,:,:,:,iws)))
else
  tmp1(:,:,:,:,ith) = tmp1(:,:,:,:,ith) +  wfld(:,:,:,:,iws)
end if

end subroutine
subroutine grab_wavefield(wfld,iws,ith)
complex :: wfld(:,:,:,:,:)
integer :: ith,iws,i1,i2,ierr
if(run%mem_save) then
  ierr= sseek(twave,0,0)
  ierr= sreed(twave,wfld(:,:,:,:,iws),4*size(wfld(:,:,:,:,iws)))
 else 
  wfld(:,:,:,:,iws) = tmp1(:,:,:,:,ith)
end if
end subroutine

subroutine store_part_wavefield(wfld,ifk,ith,iws,izs)
 complex wfld(:,:,:,:,:)
  integer :: ifk,ith,iws,ierr,izs,i1,i2
if(run%mem_save) then
  call partition_slow(wfld(:,:,:,:,iws),ifk,izs)
  ierr= sseek(twave,0,0)
   do i2=1,size(wfld,4)
     do i1=1,size(wfld,3)
       ierr= sreed(twave,tmps,size(tmps)*4)
       wfld(:,:,i1,i2,iws)=wfld(:,:,i1,i2,iws)+tmps
     end do
   end do 
  ierr= sseek(twave,0,0)
  ierr= srite(twave,wfld(:,:,:,:,iws),4*size(wfld(:,:,:,:,iws)))
else
   where(slo%si(:,:,:,:,izs) == ifk)
      tmp1(:,:,:,:,ith) = tmp1(:,:,:,:,ith) + wfld(:,:,:,:,iws)
   end where
end if
end subroutine
end module
