module wei_mig
  !! WEI migration

  use wei_util
  use wei_run
   
  use wei_obj_S
  use wei_obj_R
  use wei_obj_D
  use wei_obj_W

  use wei_fft
  use wei_kmap
  use wei_xmap

  use wei_kxmig
  use wei_image
  !---------------
  implicit none

  logical, private :: wIO
  type(wedata) :: D,bW
  type(weimag) :: R


contains
!!$
!!$=head1 
!!$
!!$ weimig_init - Initialize migration operator
!!$
!!$=head1 SYNOPSIS
!!$
!!$  call weimig_init(SLin,WCin,FKin,FXin,IGin,in6,in7,in8,in9)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item SLin - Slowness operator init  L<weislo1_init>
!!$
!!$=item WCin - WCop operator init  L<weikxmig1_init>
!!$
!!$=item FKin - FK operator init  L<weiwem_init>
!!$
!!$=item FXin - FX operator init  L<weissf_init>
!!$
!!$=item IGin - image gather creation operator init  L<weihcig_init>
!!$
!!$=item in6,..in9 -  integer,optional
!!$
!!$ Dummy operators for future expansion
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Initialize migration/modeling operator
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<weimig> L<wei>
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=cut
!!$
  !----------------------------------------------------------------
  subroutine weimig_init(SLin,WCin,FKin,FXin,IGin,in6,in7,in8,in9)
    integer,external,optional ::                     in6,in7,in8,in9
    integer :: st,er
    character(len=128) :: name
#ifdef DEBUG
    character(len=128), parameter :: unit='weimig_init'
#endif
    interface !!-------------------------
       integer function    SLin() result(st)
       end function SLin
       integer function WCin() result(st)
       end function WCin
       integer function    FKin() result(st)
       end function FKin
       integer function    FXin() result(st)
       end function FXin
       integer function    IGin() result(st)
       end function IGin
    end interface !!---------------------

    !! initialize the main WEI variables
    if(.not. weirun_init()) &
      call erexit("trouble initializing weirun_init")
    if(.not. down_init()) &
      call erexit("trouble initializing down_init")
    
#ifdef DEBUG
    if(run%debug) call in(unit)
#endif

    !! flag whether I/O on wavefield.
    wIO=.false.

    !! initialize objects
    if(.not. wei_slow_init(S,down%VS,down%VG))&
      call erexit("trouble initializing slowness")
    if(.not. init_distrib_pattern()) &
      call erexit("trouble initializing distribution pattern")
    if(.not. wei_image_init(R,down%R)) &
      call erexit("trouble initializing image")
    if(.not. wei_data_init(D,down%D)) &
      call erexit("trouble initializing data")


    !! auxiliary objects
    name="bwfld.WEI"  
    if(.not. wei_aux_data_init(bW,name))&   !CREATE AN AUXILARY WAVEFIELD FILE
      call erexit("trouble initializing wavefield object")
    
    if(wIO) call weibwfld_init(W,down%W) !! needs aux obj  

    down%mod_ = down%R   !MODEL IS REFLECTIVITY
    down%dat_ = down%D   !DATA IS FREQUENCY SLICES

    !! convert frequency to radians
    call process_radians()

    !! init FFT
    call weifft_init()  

    !! init K-map
    call weikmap_init()
    call weixmap_init()

    !! init imaging
    call weiimage_init()

    !! allocate slowness arrays for FK/FX ops
    call weislow_allocate()

    !! check if offset is consistent with ph
    if(down%pCIG) call check_hp()

    !! init operators
    st=SLin()     !! FK slownesses init
    st=WCin()  !! WCop init
    st=FKin()     !! FK op init
    st=FXin()     !! FX op init
    st=IGin()     !! FX op init

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weimig_init
  !----------------------------------------------------------------
  !----------------------------------------------------------------
  subroutine weimig_outnull()
  !ZERO OUTPUT IF REQUESTED
#ifdef DEBUG
    character(len=128), parameter :: unit='weimig_outnull'
    if(run%debug) call in(unit)
#endif

    if(down%adj) then !! R<-D
       call     image_assign(R,(0.,0.))
    else             !! R->D
       call      data_assign(D,(0.,0.))
    end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weimig_outnull
  !----------------------------------------------------------------
!!$
!!$=head1 
!!$
!!$ weimig -  Migration/modeling operator
!!$
!!$=head1 SYNOPSIS
!!$
!!$  ierr=weimig(adj,add,mod,dat,SLop,WCop,FKop,FXop,IGop,in6,in7,in8,in9)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item adj,add - logcal
!!$
!!$ Adjoint and add (adjoint=.true. - migration)
!!$ (adjoint=.false. - modeling).  add=.true. add to current file
!!$
!!$=item mod,dat - char*
!!$
!!$ Tags to the model and data
!!$
!!$=item SLop - Slowness operator   L<weislo1>
!!$
!!$=item WCop - WCop operator   L<weikxmig1>
!!$
!!$=item FKop - FK operator   L<weiwem>
!!$
!!$=item FXop - FX operator   L<weissf>
!!$
!!$=item IGop - image gather creation operator   L<weihcig>
!!$
!!$=item in6,..in9 -  integer,optional
!!$
!!$ Dummy operators for future expansion
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Migration/modeling operator
!!$
!!$ Three possible schemes  
!!$
!!$ Outer frequency---
!!$
!!$ iw_blocks{
!!$
!!$   iz_block{
!!$
!!$     migration
!!$
!!$   }
!!$
!!$ }
!!$
!!$ Outer depth---
!!$
!!$ iz_block{
!!$
!!$   iw_blocks{
!!$
!!$     migration
!!$
!!$   }
!!$
!!$ }
!!$
!!$ Outer frequency, variable steps---
!!$
!!$ iw_blocks{
!!$
!!$   iz_block (function of freq/slowness){
!!$
!!$     migration
!!$
!!$   }
!!$
!!$ }
!!$
!!$ If distribute specified 
!!$
!!$ distribute=vsf - variable steps as a function of frequency and slowness
!!$
!!$ distribute=vs - variable steps as a function of slowness
!!$
!!$ distribute=reg -  regular sampling methodology determination:
!!$
!!$ If outer_freq set then used otherwise
!!$
!!$ If few frequency blocks  outer frequency used
!!$
!!$ If few depth blocks  outer depth used
!!$
!!$
!!$=head1 SEE ALSO
!!$
!!$ L<weimig_init>
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=cut
!!$
  !----------------------------------------------------------------
  integer function weimig(adj,add,mod_,dat_,SLop,  WCop,FKop,FXop,IGop,op6,op7,op8,op9)
  logical,            intent(in) :: adj,add
  character(len=128), intent(in) :: mod_,dat_
    integer,external,optional :: WCop,FKop,FXop,IGop,op6,op7,op8,op9
    integer :: izb,iwb
    integer :: st
     complex, allocatable,target :: buffer(:,:,:,:,:)
#ifdef DEBUG
    character(len=128), parameter :: unit='weimig'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%debug) call in(unit)
#endif

    !! reset tag variables
    call down_reset(adj,add,mod_,dat_)
!    if(run%verb) call process_follow() !! where am I?
    
    !! accomodate tag name changes
#ifdef SEP_MPI
    R%tag_0 = down%mod_
    D%tag_0 = down%dat_
#else
    R%tag = trim(down%mod_)
    D%tag = trim(down%dat_)
#endif
   


    !! setup the adjoint sign
    if(down%adj) then
       down%opsign = goT2B
    else
       down%opsign = goB2T
    end if

    if(run%mem_save) then
      deallocate(down%block)
      allocate(buffer(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,1))
      bW%bin=>buffer
      D%bin=>buffer
      R%bin=>buffer
      D%bin=>buffer
    end if

!    !! zero the output if add is off
    if(.not. down%add) call weimig_outnull()

!    call to_history("nz",down%nzb,R%tag) 
    if (run%mem_save) then
       weimig=weimigsave(SLop,  WCop,FKop,FXop,IGop)
    else if(down%order(1:3)=="VAR") then
       weimig=weimigvar(SLop,  WCop,FKop,FXop,IGop)
    else if(down%order(1:2)=="OF") then
       weimig=weimigof(SLop,  WCop,FKop,FXop,IGop)
    else 
       weimig=weimigoz(SLop,  WCop,FKop,FXop,IGop)
    end if

#ifdef SEP_MPI
    if(run%verb) call weimessage("COMBINE DATA")
    call combine_data(D)

    if(run%verb) call weimessage("COMBINE IMAGE")
    call combine_image(R)

    if(run%verb) call weimessage("COMBINE SLOW")
!    call combine_slow(S)

    if(run%verb) call weimessage("OK")

    call aux_unlink(trim(bW%tag))
if(down%order(1:3) == "VAR") call aux_unlink(S%s_s%tag_var)
    call aux_unlink(S%s_s%tag)
    if(down%conv_wave) call aux_unlink(S%s_g%tag)



#endif

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    weimig=OK
    
  end function weimig

  !----------------------------------------------------------
  integer function weimigvar(SLop,WCop,FKop,FXop,IGop,op6,op7,op8,op9)
    integer,external,optional :: WCop,FKop,FXop,IGop,op6,op7,op8,op9
    integer :: izb,iwb,izdone,nzs,nzb
    integer :: st
#ifdef DEBUG
    character(len=128), parameter :: unit='weimigvar'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------

    
    izb=0
    do iwb=1,size(var%z_samp)
      izb=izb+ceiling(size(var%z_samp(iwb)%dz)/(1.0*down%nzs))
    end do
    call set_blocks(n_wb=size(down%iw_first),n_blocks=izb)


    if(down%adj) then !! MIGRATION
#ifdef DEBUG
       if(run%debug) call in('migration')
#endif
       !----------------------------------------------------------
       call data_copy(D,bW)           !! copy the data at the top
!!          call to_history("fblock",0,R%tag) 
       wbA: do iwb=1,down%nwb
          if(.not. data_read(bW,iwb))&      !! read  data at block top
             call erexit("trouble reading data")
          call var_samp_iw(iwb)
          call var_slow(iwb)
          call set_blocks(n_zb=ceiling(1.0*size(var%z_samp(iwb)%dz)/(1.0*down%nzs)))
          izdone=0;izb=0
          nzb=ceiling(real(size(var%z_samp(iwb)%dz))/real(down%nzs))
          zbA: do while(izdone< size(var%z_samp(iwb)%dz)) 
            izb=izb+1
            call setup_var_block(iwb,izb)
            nzs=min(down%nzs,size(var%z_samp(iwb)%dz)-izdone)
            R%bin=cmplx(0.,0.)
            call slowness_read(S,izb,nzs) !! read a slowness block
            st = SLop()               !! setup sS, sR, s(z)
            if(run%verb) call wei_followup(izb,iwb)
            call wei_followup(izb,iwb)

            down%bw%o = down%aw%o + (down%iw_first(iwb)-1)*down%bw%d
            if(run%mem_save) &
            call image_write(R,izb,iwb/=1.or.down%add .or. run%share_image,iwb,nzs)!!write image block
            call wmigRD(izb,iwb,nzs,WCop,FKop,FXop,IGop) !! trash W
            if(.not. run%mem_save) &
            call image_write(R,izb,iwb/=1.or.down%add .or. run%share_image,iwb,nzs)!!write image block
            izdone=izdone+nzs

          end do zbA
!          call to_history("fblock",iwb,R%tag) 
       end do wbA
       !----------------------------------------------------------
#ifdef DEBUG
       if(run%debug) call out('migration')
#endif
    else         !! MODELING

#ifdef DEBUG
       if(run%debug) call in('modeling')
#endif
       call seperr("unimplemented variable sampling modeling")
#ifdef DEBUG
       if(run%debug) call out('modeling')
#endif
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    weimigvar=OK
  end function weimigvar
  !----------------------------------------------------------
  integer function weimigsave(SLop,WCop,FKop,FXop,IGop,op6,op7,op8,op9)
    integer,external,optional :: WCop,FKop,FXop,IGop,op6,op7,op8,op9
    integer :: izb,iwb,iws,izs,ith
    integer :: st
#ifdef DEBUG
    character(len=128), parameter :: unit='weimigsave'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------
     var%dzb=down%az%d


#ifdef DEBUG
       if(run%debug) call in('migration')
#endif
       !----------------------------------------------------------
       call set_blocks(n_wb=size(down%iw_first),n_zb=down%az%n/down%nzs,n_blocks=down%nwb*down%nzb)
       call data_copy(D,bW)           !! copy the data at the top
       call to_history("fblock",0,R%tag) 
       wbA: do iwb=1,down%nwb
          if(.not. data_read(bW,iwb))&      !! read  data at block top
            call erexit("trouble reading data")
          zbA: do izb=1,down%nzb       
            call setup_var_block(iwb,izb)
            call slowness_read(S,izb) !! read a slowness block
            st = SLop()               !! setup sS, sR, s(z)
            if(run%verb) call wei_followup(izb,iwb)
            if(.not. run%mem_save) R%bin=0.;
            down%bw%o = down%aw%o + (down%iw_first(iwb)-1)*down%bw%d
            call store_w(bw%bin)
            call weiimgW2R(1,1,1,R%bin,bW%bin,IGop)
            call image_write(R,izb,(iwb/=1 .or. down%add .or. run%share_image)) !! write image block
            call grab_w(bw%bin)
            call weikxmig(bW%bin,1,1,1,WCop,FKop,FXop)
          end do zbA
          call to_history("fblock",iwb,R%tag) 
       end do wbA
       !----------------------------------------------------------
#ifdef DEBUG
       if(run%debug) call out('migration')
#endif

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    weimigsave=OK
  end function weimigsave
!--------------------------------------------------------------------
!--------------------------------------------------------------------
  integer function weimigof(SLop,WCop,FKop,FXop,IGop,op6,op7,op8,op9)
    integer,external,optional :: WCop,FKop,FXop,IGop,op6,op7,op8,op9
    integer :: izb,iwb
    integer :: st,ierr
#ifdef DEBUG
    character(len=128), parameter :: unit='weimigof'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------
     var%dzb=down%az%d


    if(down%adj) then !! MIGRATION
#ifdef DEBUG
       if(run%debug) call in('migration')
#endif
       !----------------------------------------------------------
       call set_blocks(n_wb=size(down%iw_first),n_zb=down%az%n/down%nzs,n_blocks=down%nwb*down%nzb)
       call data_copy(D,bW)           !! copy the data at the top
!       call to_history("fblock",0,R%tag) 
       wbA: do iwb=1,down%nwb
          if(.not. data_read(bW,iwb))&      !! read  data at block top
            call erexit("trouble reading data")
          zbA: do izb=1,down%nzb       
            call setup_var_block(iwb,izb)
            call slowness_read(S,izb) !! read a slowness block
            st = SLop()               !! setup sS, sR, s(z)
            if(run%verb) call wei_followup(izb,iwb)
            R%bin=0.;
            down%bw%o = down%aw%o + (down%iw_first(iwb)-1)*down%bw%d
            if(wIO) then           !! DC one block
                call wmigWW(izb,iwb,WCop,FKop,FXop,IGop) !! write W
             else
                call wmigRD(izb,iwb,down%nzs,WCop,FKop,FXop,IGop) !! trash W
             end if
            call image_write(R,izb,(iwb/=1 .or. down%add .or. run%share_image)) !! write image block
          end do zbA
!          call to_history("fblock",iwb,R%tag) 
       end do wbA
       !----------------------------------------------------------
#ifdef DEBUG
       if(run%debug) call out('migration')
#endif
    else         !! MODELING
#ifdef DEBUG
       if(run%debug) call in('modeling')
#endif
       !----------------------------------------------------------
       call data_assign(bW,(0.,0.))    !! start with no data
       call set_blocks(n_wb=size(down%iw_first),n_zb=down%az%n/down%nzs,n_blocks=down%nwb*down%nzb)
       wbF: do iwb=1,down%nwb
           if(.not. data_read(bW,iwb))&      !! read  data at block bottom
             call erexit("trouble reading data")
           zbF: do izb=down%nzb,1,-1    
            call setup_var_block(iwb,izb)
             call    image_read(R,izb) !! read image block
             call slowness_read(S,izb) !! read a slowness block
             st = SLop()               !! setup sS,sR, s(z)
             if(run%verb) call wei_followup(izb,iwb)
!             down%bw%o = down%aw%o + (iwb-1)*(down%bw%n*down%bw%d)
             down%bw%o = down%aw%o + (down%iw_first(iwb)-1)*down%bw%d
             if(wIO) then           !! DC one block
                call wmigWW(izb,iwb,WCop,FKop,FXop,IGop) !! write W
             else
                call wmigRD(izb,iwb,down%nzs,WCop,FKop,FXop,IGop) !! trash W
             end if
            end do zbF
            if(.not. data_write(bW,iwb))& !! write data at block top
              call erexit("trouble writing data")
!             call to_history("fblock",iwb,D%tag) 
        end do wbF
       call data_add(bW,D)          !! add U.C. wfld to data
       !----------------------------------------------------------
#ifdef DEBUG
       if(run%debug) call out('modeling')
#endif
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    weimigof=OK
  end function weimigof


  !----------------------------------------------------------
  integer function weimigoz(SLop,WCop,FKop,FXop,IGop,op6,op7,op8,op9)
    integer,external,optional :: WCop,FKop,FXop,IGop,op6,op7,op8,op9
    integer :: izb,iwb
    integer :: st
#ifdef DEBUG
    character(len=128), parameter :: unit='weimigoz'
#endif
    interface !!----------------------------------------------
       integer function SLop() result(st)
       end function SLop
    end interface !!------------------------------------------

       var%dzb=down%az%d
       call set_blocks(n_wb=size(down%iw_first),n_zb=down%az%n,n_blocks=down%nwb*down%nzb)
    if(down%adj) then !! MIGRATION
#ifdef DEBUG
       if(run%debug) call in('migration')
#endif
       !----------------------------------------------------------
       call data_copy(D,bW)         !! copy the data at the top
       zbA: do izb=1,down%nzb        
          call    image_read(R,izb) !! read image block
          call slowness_read(S,izb) !! read a slowness block
          st = SLop()               !! setup sS, sR, s(z)
!          call to_history("fblock",0,R%tag) 
          wbA: do iwb=1,down%nwb
            call setup_var_block(iwb,izb)
            if(run%verb) call wei_followup(izb,iwb)
             down%bw%o = down%aw%o + (down%iw_first(iwb)-1)*down%bw%d
             if(.not. data_read(bW,iwb))& !! read  data at block top
               call erexit("trouble reading data")
             if(wIO) then           !! DC one block
                call wmigWW(izb,iwb,WCop,FKop,FXop,IGop) !! write W
             else
                call wmigRD(izb,iwb,down%nzs,WCop,FKop,FXop,IGop) !! trash W
             end if
             if(.not. data_write(bW,iwb))&  !! write data at block bottom
              call erexit("trouble writing data")
             call to_history("fblock",iwb,R%tag) 
          end do wbA
            call image_write(R,izb)     !! write image block
       end do zbA
       !----------------------------------------------------------
#ifdef DEBUG
       if(run%debug) call out('migration')
#endif
    else         !! MODELING
#ifdef DEBUG
       if(run%debug) call in('modeling')
#endif
       !----------------------------------------------------------
       call data_assign(bW,(0.,0.)) !! start with no data
       zbF: do izb=down%nzb,1,-1
          call    image_read(R,izb) !! read image block
          call slowness_read(S,izb) !! read a slowness block
          st = SLop()               !! setup sS,sR, s(z)
!          call to_history("fblock",0,D%tag) 
          wbF: do iwb=1,down%nwb
            call setup_var_block(iwb,izb)
             if(run%verb) call wei_followup(izb,iwb)
!             down%bw%o = down%aw%o + (iwb-1)*(down%bw%n*down%bw%d)
             down%bw%o = down%aw%o + (down%iw_first(iwb)-1)*down%bw%d
             if(.not. data_read(bW,iwb))& !! read  data at block bottom
               call erexit("trouble reading data")
             if(wIO) then           !! DC one block
                call wmigWW(izb,iwb,WCop,FKop,FXop,IGop) !! write W
             else
                call wmigRD(izb,iwb,down%nzs,WCop,FKop,FXop,IGop) !! trash W
             end if
             if(.not. data_write(bW,iwb))&!! write data at block top
              call erexit("trouble writing data")
!             call to_history("fblock",iwb,D%tag) 
          end do wbF
       end do zbF
       call data_add(bW,D)          !! add U.C. wfld to data
       !----------------------------------------------------------
#ifdef DEBUG
       if(run%debug) call out('modeling')
#endif
    end if
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    weimigoz=OK
  end function weimigoz
  !---------------------------------------------------------------
  !! don't write the wavefield
  subroutine wmigRD(izb,iwb,nzs,WCop,FKop,FXop,IGop)
    integer, intent(in)       :: izb,iwb
    integer,external,optional :: WCop,FKop,FXop,IGop
    integer :: izs,iws,ith,nzs,ierr,i
#ifdef DEBUG
    character(len=128), parameter :: unit='wmigRD'
    if(run%debug) call in(unit)
#endif

!            if(run%m_ith==0) write(0,*) "READ1CHECK",sum(cabs(bw%bin)),sum(cabs(bw%bin(:,:,:,:,1)))

    if(down%adj) then !! MIGRATION
#ifdef SEP_OMP
       !$OMP  PARALLEL
       !$OMP  DO &
       !$OMP& SCHEDULE(DYNAMIC) &
       !$OMP& PRIVATE(iws,izs,ith)
#endif
       wsA: do iws=1,down%nws             !LOOP OVER w STEPS
#ifdef SEP_OMP
          ith = omp_get_thread_num()+1
#else
          ith=1
#endif
          zsA: do izs=1,nzs          !LOOP OVER z STEPS
!          if(run%m_ith==2 .and. ith==1) write(0,*) "BEFORE 0",izs,iws,sum(cabs(bw%bin(:,:,:,:,iws)))
             call weiimgW2R(izs,iws,ith,R%bin,bW%bin,IGop)
!          if(run%m_ith==2 .and. ith==1) write(0,*) "BEFORE 1",izs,iws,sum(cabs(bw%bin(:,:,:,:,iws)))
             call weikxmig(bW%bin,iws,izs,ith,WCop,FKop,FXop)
!          if(run%m_ith==2 .and. ith==1) write(0,*) "BEFORE 2",izs,iws,sum(cabs(bw%bin(:,:,:,:,iws)))
          end do zsA
       end do wsA
#ifdef SEP_OMP
       !$OMP END DO
       !$OMP END PARALLEL
#endif
    else         !! MODELLING
#ifdef SEP_OMP
       !$OMP  PARALLEL
       !$OMP  DO &
       !$OMP& SCHEDULE(DYNAMIC) &
       !$OMP& PRIVATE(iws,izs,ith)
#endif
       wsF: do iws=1,down%nws             !LOOP OVER w STEPS
#ifdef SEP_OMP
          ith = omp_get_thread_num()+1
#else
ith=1
#endif
          zsF: do izs=down%nzs,1,-1       !LOOP OVER z STEPS

             call weikxmig(bW%bin,iws,izs,ith,WCop,FKop,FXop)
             call weiimgR2W(izs,iws,ith, R%bin,bW%bin,IGop)

          end do zsF
       end do wsF
#ifdef SEP_OMP
       !$OMP END DO
       !$OMP END PARALLEL
#endif
    end if


#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine wmigRD
  !----------------------------------------------------------------
  !! write the wavefield
  subroutine wmigWW(izb,iwb,WCop,FKop,FXop,IGop)
    integer, intent(in)       :: izb,iwb
    integer,external,optional :: WCop,FKop,FXop,IGop
    integer :: izs,iws,ith
#ifdef DEBUG
    character(len=128), parameter :: unit='wmigWW'
    if(run%debug) call in(unit)
#endif

    if(down%adj) then !! MIGRATION
       !$OMP  PARALLEL
       !$OMP  DO &
       !$OMP& SCHEDULE(DYNAMIC) &
       !$OMP& PRIVATE(iws,izs,ith)
       wsA: do iws=1,down%nws
#ifdef SEP_OMP
          ith = omp_get_thread_num()+1
#else
          ith=1
#endif
          zsA: do izs=1,down%nzs

             !$OMP CRITICAL
             call wavefield_write(W,bW%bin,izb,izs,iwb,iws)
             !$OMP END CRITICAL

             call weiimgW2R(izs,iws,ith, R%bin,bW%bin,IGop)
             call weikxmig(bW%bin,iws,izs,ith,WCop,FKop,FXop)

          end do zsA
       end do wsA
       !$OMP END DO
       !$OMP END PARALLEL
    else         !! MODELLING
       !$OMP  PARALLEL
       !$OMP  DO &
       !$OMP& SCHEDULE(DYNAMIC) &
       !$OMP& PRIVATE(iws,izs,ith)
       wsF: do iws=1,down%nws
#ifdef SEP_OMP
          ith = omp_get_thread_num()+1
#else
          ith = 1
#endif
          zsF: do izs=down%nzs,1,-1

             call weikxmig(bW%bin,iws,izs,ith,WCop,FKop,FXop)
             call weiimgR2W(izs,iws,ith, R%bin,bW%bin,IGop)

             !$OMP CRITICAL
             call wavefield_write(W,bW%bin,izb,izs,iwb,iws)
             !$OMP END CRITICAL

          end do zsF
       end do wsF
       !$OMP END DO
       !$OMP END PARALLEL
    end if

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine wmigWW
  !----------------------------------------------------------
end module wei_mig


!!   ADJ         FOR
!!
!! D1 > R1     D1 > R1
!! v           ^
!! D2 > R2     D2 > R2
!! v           ^
!! D3 > R3     D3 > R3
!! v           ^
!! D4          D4
!! v           ^
!! ========    ========
!! v           ^
!! D4 > R4     D4 > R4
!! v           ^
!! D5 > R5     D5 > R5
!! v           ^
!! D6 > R6     D6 > R6
!! v           ^
!! D           D
!!



