module wei_ssf
  !! Split-Step Fourier

  use wei_util
  use wei_run
  use wei_obj_S
  !---------------
  implicit none

contains
  !----------------------------------------------------------------
!!$
!!$=head1 NAME
!!$
!!$ weissf_init - Initialize split step operator
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr=weissf_init()
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Initialize split step operator
!!$
!!$=head1 SEE ALSO
!!$
!!$L<weissf>
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=cut
!!$
  integer function weissf_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weissf_init'
    if(run%debug) call in(unit)
#endif
    !! do nothing
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weissf_init
  !----------------------------------------------------------------
!!$
!!$=head1 NAME
!!$
!!$ weissf - Split step operator
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr=weissf_init(iws,izs,ifk,ith,wfld)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item iws - integer
!!$
!!$  Current frequency slice
!!$
!!$=item izs - integer
!!$
!!$  Current depth slice
!!$
!!$=item ifk - integer
!!$
!!$  Current fk operation
!!$
!!$=item ith - integer
!!$
!!$  Current thread
!!$
!!$=item wfld - complex(:,:,:,:,:)
!!$
!!$ Wavefield
!!$
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Split step operator. Apply first order slowness correction
!!$  to wthe wavefield given difference between velocity at
!!$  location and reference velocity.
!!$
!!$=head1 SEE ALSO
!!$
!!$L<weissf_init>
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=cut
!!$
!! Split-step
  integer function weissf(iws,izs,ifk,ith,wfld) result(st)
    integer, intent(in) :: iws,izs,ifk,ith
    complex, dimension(:,:,:,:,:), pointer :: wfld
    real    ::  w
    integer :: i1,i2,i3,i4
    complex :: iwdz
#ifdef DEBUG
    character(len=128), parameter :: unit='weissf'
    if(run%depar) call in(unit)
#endif
    w    = down%bw%o + (iws-1)*down%bw%d
    iwdz = (0,1)* down%opsign*w*var%dzb(izs)


    if(.not. run%mem_save) then
      if(slo%th(ith)%source) then
        do i4=1,size(wfld,4);do i3=1,size(wfld,3);
         do i2=1,size(wfld,2);do i1=1,size(wfld,1)
          wfld(i1,i2,i3,i4,iws) = &
          wfld(i1,i2,i3,i4,iws) * exp( iwdz *(&
            ( real(S%pba(i1,i2,i3,i4,izs))-slo%sz_s(izs,ifk))))
         end do; end do; end do; end do;
        else
        do i4=1,size(wfld,4);do i3=1,size(wfld,3);
          do i2=1,size(wfld,2);do i1=1,size(wfld,1)
          wfld(i1,i2,i3,i4,iws) = &
          wfld(i1,i2,i3,i4,iws) * exp( iwdz *(&
            (aimag(S%pba(i1,i2,i3,i4,izs))-slo%sz_r(izs,ifk))  ))
         end do; end do; end do; end do;
      end if

    else
      if(slo%th(ith)%source) then
        do i4=1,size(wfld,4);do i3=1,size(wfld,3)
          call fill_small_pba(i3,i4)
          do i2=1,size(wfld,2);do i1=1,size(wfld,1)
            wfld(i1,i2,i3,i4,iws) = &
            wfld(i1,i2,i3,i4,iws) * exp( iwdz *(&
            (real(S%pba_s(i1,i2,izs))-slo%sz_s(izs,ifk))  ))
          end do; end do; end do; end do;
       else
        do i4=1,size(wfld,4);do i3=1,size(wfld,3)
          call fill_small_pba(i3,i4)
          do i2=1,size(wfld,2);do i1=1,size(wfld,1)
            wfld(i1,i2,i3,i4,iws) = &
            wfld(i1,i2,i3,i4,iws) * exp( iwdz *(&
            (aimag(S%pba_s(i1,i2,izs))-slo%sz_r(izs,ifk))  ))
          end do; end do; end do; end do;
       end if
    end if

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weissf
  !----------------------------------------------------------------
end module wei_ssf
